import datetime
import json
import os
from uuid import uuid4
import pandas as pd
import numpy as np

from slate.api import API
from slate.utils import assemble_base


class Backtest:
    def __init__(self, api: API):
        """
        Initialize a new live instance
        :param api: The object containing the Blankly API
        """
        self.__api: API = api

        self.__live_base = '/v1/backtest'

    def __assemble_base(self, route: str) -> str:
        """
        Assemble the sub-route specific to live posts

        :param route: The sub route: /spot-market -> /v1/live/spot-market
        :return: str
        """
        return assemble_base(self.__live_base, route)

    @staticmethod
    def generate_new_backtest_id() -> str:
        """
        Create a new unique backtest identifier

        :return: A string uuid which can be used for backtesting
        """

        # Import UUID here to reduce load time
        import uuid
        return str(uuid.uuid4())

    def result(self,
               symbols: list,
               quote_asset: str,
               start_time: [int, float, datetime.datetime],
               stop_time: [int, float, datetime.datetime],
               account_values: [list, pd.Series, np.ndarray],
               trades: [list, pd.Series, np.ndarray],
               exchange: str,
               backtest_id: str = None,
               metrics: dict = None,
               indicators: dict = None,
               time: datetime.datetime = None
               ) -> dict:
        """
        Post a backtest result object to the platform

        **Look at this link to learn more**:
        https://docs.blankly.finance/services/events#post-v1backtestresult
        """
        if backtest_id is None:  # generate one if they don't input one
            backtest_id = str(uuid4())

        # Convert any ndarray
        if isinstance(account_values, np.ndarray):
            account_values = account_values.tolist()
        if isinstance(trades, np.ndarray):
            trades = trades.tolist()

        if isinstance(account_values, pd.Series):
            account_values = account_values.tolist()
        if isinstance(trades, pd.Series):
            trades = trades.tolist()

        if isinstance(start_time, datetime.datetime):
            start_time = start_time.timestamp()
        if isinstance(stop_time, datetime.datetime):
            stop_time = stop_time.timestamp()

        data = {
            'symbols': symbols,
            'quote_asset': quote_asset,
            'start_time': start_time,
            'stop_time': stop_time,
            'account_values': account_values,
            'trades': trades,
            'exchange': exchange,
            'metrics': metrics,
            'backtest_id': backtest_id,
            'indicators': indicators,
        }
        files = {}
        if len(account_values) > 0:
            import tempfile
            account_values_json = {
                'account_values': account_values
            }
            fd, path = tempfile.mkstemp()
            open(path, 'w').write(json.dumps(account_values_json, indent=2))
            os.close(fd)
            files['account_values'] = path
            data.pop('account_values')
        if len(trades) > 0:
            import tempfile
            trades_json = {
                'trades': trades
            }
            fd, path = tempfile.mkstemp()
            open(path, 'w').write(json.dumps(trades_json, indent=2))
            os.close(fd)
            files['trades'] = path
            data.pop('trades')

        return self.__api.post(self.__assemble_base('/result'), data, time, files_=files)

    def status(self,
               successful: bool,
               status_summary: str,
               status_details: str,
               time_elapsed: [int, float],
               backtest_id: str,
               description: str = None,
               label: str = None,
               time: datetime.datetime = None) -> dict:
        """
        Post a backtest status. This is primarily used for backtest lifecycle
        https://docs.blankly.finance/services/events#post-v1livelog

        :param successful: A boolean specifying if it succeeds or fails
        :param status_summary: A brief message if the status has completed or not
        :param status_details: A more extensive message about the backtest
        :param time_elapsed: The amount of time taken to start and run the backtest
        :param backtest_id: The identifier for the backtest
        :param time: A time object to fill if the event occurred in the past
        :param description: Post or overwrite the backtest description
        :param label: Add a label to your backtest. Should be short.
        :return: API response (dict)
        """
        return self.__api.post(self.__assemble_base('/status'), {
            'successful': successful,
            'status_summary': status_summary,
            'status_details': status_details,
            'time_elapsed': time_elapsed,
            'backtest_id': backtest_id,
            'description': description,
            'label': label
        }, time)

    def log(self, line: str, type_: str, backtest_id: str, time: datetime.datetime = None) -> dict:
        """
        Post a new log from a backtesting model. This can be stdout or any custom logs
        https://docs.blankly.finance/services/events#post-v1backtestlog

        :param line: The line to write
        :param type_: The type of line, common types include 'stdout' or 'stderr'
        :param backtest_id: The identifier for the backtest
        :param time: A time object to fill if the event occurred in the past
        :return: API response (dict)
        """
        return self.__api.post('/log', {
            'line': line,
            'type': type_,
            'backtest_id': backtest_id
        }, time)



from slate.api import API
from slate.utils import assemble_base


class Model:
    def __init__(self, api: API):
        """
        Initialize a new live instance
        :param api: The object containing the Blankly API
        """
        self.__api: API = api

        self.__live_base = '/v1/model'

    def __assemble_base(self, route: str) -> str:
        """
        Assemble the sub-route specific to live posts

        :param route: The sub route: /spot-market -> /v1/live/spot-market
        :return: str
        """
        return assemble_base(self.__live_base, route)

    def set_lifecycle(self,
                      message: str = None,
                      start_at: [int, float] = None,
                      end_at: [int, float] = None,
                      running: bool = None) -> dict:
        """
        Update the model lifecycle
        All keys are optional so that any individual parameter can be updated
        https://docs.blankly.finance/services/events#post-v1modellifecycle

        :param message: A short message about status like 'Installing Dependencies'
        :param start_at: The start time in epoch
        :param end_at: The end time in epoch
        :param running: A boolean specifying if the model is running or not
        :return: API response (dict)
        """
        return self.__api.post(self.__assemble_base('/lifecycle'), {
            'message': message,
            'start_at': start_at,
            'end_at': end_at,
            'running': running
        })

    # TODO add routes to remove the used symbols and exchanges
    def add_symbol(self,
                   symbol: str) -> dict:
        """
        Add a used symbol for the live view to the platform

        :param symbol: A string like 'AAPL' or 'BTC-USD'
        :return: API response (dict)
        """
        return self.__api.post(self.__assemble_base('/used-symbol'), {
            'symbol': symbol
        })

    def set_exchange(self, used_exchange) -> dict:
        """
        Add an exchange to the list of used exchanges to the platform

        :param used_exchange: Exchange like 'coinbase_pro' or 'oanda'
        :return: API response (dict)
        """
        return self.__api.post(self.__assemble_base('/used-exchange'), {
            'exchange': used_exchange
        })



import json
from operator import itemgetter

import slate
from slate.integrations.common import b_id, DUMMY_METRICS, DUMMY_INDICATORS

try:
    import jesse
except ImportError:
    pass


class JesseAi:
    slate: 'slate.Slate'
    api: 'slate.api.API'

    def __init__(self, slate, api):
        self.slate = slate
        self.api = api

    def post_backtest(self, json_result: str):
        with open(json_result, 'r') as file:
            trades = json.load(file)['trades']

        trades = [trade for j_trade in trades for trade in map_trades(j_trade)]
        trades.sort(key=itemgetter('time'))

        account_values = []
        current_value = 0
        for trade in trades:
            if 'PNL' in trade:
                current_value += trade['PNL']
                account_values.append({
                    'time': trade['time'],
                    'value': current_value
                })

        symbols = list({trade['symbol'] for trade in trades})
        quote_asset = symbols[0].split('-')[1] if '-' in symbols[0] else 'USD'
        start = trades[0]['time']
        end = trades[-1]['time']
        id = b_id()

        self.slate.backtest.result(symbols=symbols,
                                   quote_asset=quote_asset,
                                   start_time=start,
                                   stop_time=end,
                                   account_values=account_values,
                                   trades=trades,
                                   backtest_id=id,
                                   metrics=DUMMY_INDICATORS,
                                   indicators=DUMMY_METRICS)

        self.slate.backtest.status(backtest_id=id,
                                   successful=True,
                                   status_summary='Completed',
                                   status_details='',
                                   time_elapsed=0)


def map_trades(trade):
    common = {
        'symbol': trade['symbol'],
        'size': trade['size'],
        'type': 'market'
    }
    open = {**common,
            'time': trade['opened_at'],
            'side': 'buy' if trade['type'] == 'long' else 'sell',
            'price': trade['entry_price'],
            'id': b_id()}
    close = {**common,
             'time': trade['closed_at'],
             'side': 'sell' if trade['type'] == 'long' else 'buy',
             'price': trade['exit_price'],
             'id': b_id(),
             'PNL': trade['PNL']}  # used to calculate account values
    return [open, close]


if __name__ == '__main__':
    # post jesse-ai backtest to slate
    slate = slate.Slate()
    slate.integrations.jesse.post_backtest('backtest.json')



import uuid


def b_id() -> str:
    return str(uuid.uuid4())


DUMMY_INDICATORS = {'dummy': {'values': [{'time': 0, 'value': 0}],
                              'display_name': 'dummy',
                              'type': 'dummy'}}

DUMMY_METRICS = {'dummy': {'value': 0,
                           'display_name': 'dummy',
                           'type': 'dummy'}}



from operator import itemgetter

import slate
from slate.integrations.common import b_id, DUMMY_METRICS, DUMMY_INDICATORS

try:
    import bt
    from bt.backtest import Result
except ImportError:
    pass


class BtPy:
    slate: 'slate.Slate'
    api: 'slate.api.API'

    def __init__(self, slate, api):
        self.slate = slate
        self.api = api

    def post_backtests(self, result: 'Result'):
        for backtest in result.backtest_list:
            self._post_backtest(result, backtest)

    def _post_backtest(self, result: 'Result', backtest: 'Backtest'):
        symbols = [sym.upper() for sym in backtest.data.columns]
        quote = 'USD'
        for symbol in symbols:
            self.slate.model.add_symbol(symbol)
            if '-' in symbol:
                quote = symbol.split('_')[1]

        account_values = result \
            .prices[backtest.name] \
            .reset_index() \
            .rename(columns={'index': 'time', backtest.name: 'value'})
        account_values['time'] = account_values['time'].map(lambda t: t.timestamp())
        account_values = account_values.sort_values('time', ascending=True)
        account_values = account_values.to_dict('records')

        trades = result \
            .get_transactions(backtest.name) \
            .apply(map_trade, axis=1, result_type='expand') \
            .to_dict('records')

        id = b_id()
        start = result.stats[backtest.name]['start'].timestamp()
        end = result.stats[backtest.name]['end'].timestamp()
        self.slate.backtest.result(symbols=symbols,
                                   quote_asset=quote,
                                   start_time=start,
                                   stop_time=end,
                                   account_values=account_values,
                                   trades=trades,
                                   backtest_id=id,
                                   metrics=DUMMY_INDICATORS,
                                   indicators=DUMMY_METRICS)

        self.slate.backtest.status(backtest_id=id,
                                   successful=True,
                                   status_summary='Completed',
                                   status_details='',
                                   time_elapsed=0)


def map_trade(row):
    time, symbol = row.name
    return {'time': time.timestamp(),
            'side': 'buy' if row['quantity'] > 0 else 'sell',
            'symbol': symbol.upper(),
            'price': row['price'],
            'size': abs(row['quantity']),
            'id': b_id(),
            'type': 'market'}


if __name__ == '__main__':
    # run bt.py backtest
    from bt import Strategy, Backtest

    data = bt.get('spy,agg', start='2010-01-01')
    strategy = Strategy('s1', [bt.algos.RunMonthly(),
                               bt.algos.SelectAll(),
                               bt.algos.WeighEqually(),
                               bt.algos.Rebalance()])
    backtest = Backtest(strategy, data)
    result = bt.run(backtest)

    # post to slate
    slate = slate.Slate()
    slate.integrations.bt.post_backtests(result)



import os
import tempfile
import sys
import zipfile


def get_python_version():
    return '.'.join(str(i) for i in sys.version_info[:2])


def zip_dir(path, ignore_files):
    global temporary_zip_file
    temporary_zip_file = tempfile.TemporaryDirectory()
    dist_directory = temporary_zip_file.__enter__()

    # dist_directory = tempfile.TemporaryDirectory().__enter__()
    # with tempfile.TemporaryDirectory() as dist_directory:
    source = os.path.abspath(path)

    model_path = os.path.join(dist_directory, 'model.zip')
    zip_ = zipfile.ZipFile(model_path, 'w', zipfile.ZIP_DEFLATED)
    zipdir(source, zip_, ignore_files)
    zip_.close()

    return model_path


def zipdir(path, ziph, ignore_files: list):
    # From https://stackoverflow.com/a/1855118/8087739
    # Notice we're using this instead of shutil because it allows customization such as passwords and skipping
    # directories
    # ziph is zipfile handle

    filtered_ignore_files = []
    # Set all the ignored files to be absolute
    for i in range(len(ignore_files)):
        # Reject this case
        if ignore_files[i] == "":
            continue
        filtered_ignore_files.append(os.path.abspath(ignore_files[i]))

    for root, dirs, files in os.walk(path, topdown=True):
        for file in files:
            # (Modification) Skip everything that is in the blankly_dist folder
            filepath = os.path.join(root, file)

            if not (os.path.abspath(filepath) in filtered_ignore_files) and not (root in filtered_ignore_files):
                # This takes of the first part of the relative path and replaces it with /model/
                relpath = os.path.relpath(filepath,
                                          os.path.join(path, '..'))
                relpath = os.path.normpath(relpath).split(os.sep)
                relpath[0] = os.sep + 'model'
                relpath = os.path.join(*relpath)

                ziph.write(filepath, relpath)



import copy
import time
import requests
from slate.exceptions import APIException


class API:
    def __init__(self, model_id, api_key, api_pass):
        """
        Initialize the lower API class to handle requests

        :param model_id: The model id to associate the model with
        :param api_key: The API key for this project
        :param api_pass: The API pass for this project
        """

        self.__headers = {
            'model_id': model_id,
            'api_key': api_key,
            'api_pass': api_pass,
            'time': str(time.time())
        }

        # This is none for live but a datetime when set
        self.time_setting = None

        self.__api_url = 'https://events.blankly.finance'
        self.__api_version = 'v1'

    def __assemble_route_components(self, components: list) -> str:
        """
        Create a list of components that are assembled into one usable API url
        :param components: Ex: ['backtest', 'status'] -> https://events.blankly.finance/v1/backtest/status
        :return: str
        """
        url = self.__api_url + '/' + self.__api_version
        for i in components:
            url += '/'
            url += i

        return url

    def __assemble_route(self, route: str) -> str:
        """
        Append the route to the base url. This just pulls the two strings together
        :param route: route='/v1/backtest/status' -> https://events.blankly.finance/v1/backtest/status
        :return: str
        """
        return self.__api_url + route

    def __update_time(self, time_) -> dict:
        """
        Update the time in the headers

        :return: None
        """
        headers = copy.copy(self.__headers)
        if time_ is None:
            headers['time'] = str(time.time())
        else:
            headers['time'] = str(time_.timestamp())
        return headers

    @staticmethod
    def __check_errors(response: requests.Response) -> dict:
        # Check if the body is empty
        if response.raw._body is None:
            body = {}
        else:
            body = response.json()

        # Now if there is an error in the non-empty body throw an error
        # If not just return out
        if 'error' in body:
            raise APIException(body['error'])
        else:
            return body

    def post(self, route, data: dict, time_=None, files_: dict = None):
        """
        Make a basic POST request at a route
        :param route: The route without the base /v1/backtest/status
        :param data: The data to post as a dictionary in the body
        :param time_: A datetime to pass into the function
        :param files_: A dictionary containing key/values of files to file paths
        :return: dict (exchange response)
        """
        if files_ is not None:
            for i in files_:
                files_[i] = open(files_[i], 'rb')
        else:
            files_ = {}

        route = self.__assemble_route(route)
        headers = self.__update_time(time_)
        response = requests.post(route, data=data, headers=headers, files=files_)
        return response

    def get(self, route, time_=None):
        """
        Make a basic GET request at the given route
        :param route: The route without the base /time
        :param time_: A datetime to pass into the function
        :return: dict (the exchange response)
        """
        route = self.__assemble_route(route)
        headers = self.__update_time(time_)
        return requests.get(route, headers=headers)



from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, './README.md'), encoding='utf-8') as f:
    long_description = f.read()

"""
Build Info:
python3 -m build
twine upload dist/*
"""


setup(
    name='blankly-slate',
    packages=find_packages(),
    version='v1.10.2-beta',
    license='mit',
    description='View, manage and share your model from any codebase with slate',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='blankly',
    url='https://blankly.finance',
    entry_points={'console_scripts': ['slate = slate.cli:main']},
    install_requires=[
        'questionary >= 1.10.0',
        'yaspin >= 2.1.0',
        'requests',
        'pandas',
        'numpy',
    ],
    classifiers=[
        # Possible: "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
)



"""
    Updated CLI for interacting with and uploading slate models.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import json
import os.path

import questionary
import sys
import traceback
import webbrowser
from pathlib import Path
from typing import Optional
import pkgutil

from questionary import Choice

from slate.cli.api import API, blankly_deployment_url
from slate.cli.deploy import zip_dir, get_python_version
from slate.cli.login import logout, poll_login, get_token
from slate.cli.ui import text, confirm, print_work, print_failure, print_success, select, show_spinner

AUTH_URL = 'https://app.blankly.finance/auth/signin?redirectUrl=/deploy'


def validate_non_empty(text):
    if not text.strip():
        return 'Please enter a value'
    return True


def create_model(api, name, description, model_type, project_id=None):
    with show_spinner('Creating model') as spinner:
        try:
            model = api.create_model(project_id or api.user_id, model_type, name, description)
        except Exception:
            spinner.fail('Failed to create model')
            raise
        spinner.ok('Created model')
    return model


def ensure_login() -> API:
    # TODO print selected team ?
    api = is_logged_in()
    if api:
        return api
    return launch_login_flow()


def is_logged_in() -> Optional[API]:
    token = get_token()
    if not token:
        return

    # log into deployment api
    try:
        return API(token)
    except Exception:  # TODO
        return


def launch_login_flow() -> API:
    try:
        webbrowser.open_new(AUTH_URL)
        print_work(f'Your browser was opened to {AUTH_URL}. Open the window and login.')
    except Exception:
        print_work(f'Could not find a browser to open. Navigate to {AUTH_URL} and login')

    api = None
    with show_spinner(f'Waiting for login') as spinner:
        try:
            api = poll_login()
        except Exception:
            pass  # we just check for api being valid, poll_login can return None

        if not api:
            spinner.fail('Failed to login')
            sys.exit(1)

        spinner.ok('Logged in')
    return api


def slate_init(args):
    dir_is_empty = len([f for f in os.listdir() if not f.startswith('.')]) == 0
    api = ensure_login()
    model = get_model_interactive(api)

    if dir_is_empty:
        script_path = './main.py'
    else:
        script_path = questionary.path('What is the main script or entry point? (ex: main.py, bot.py)').unsafe_ask()

    with show_spinner('Generating files') as spinner:
        files = [
            ('slate.json', generate_slate_json(model, api), False),
            ('blankly.json', generate_blankly_json(model, script_path), False),
            (script_path, "if __name__ == '__main__':\n    print('Hello, World!')\n", True),
            ('requirements.txt', 'slate\n', True),
        ]
        spinner.ok('Generated files')

    for path, data, skip_existing in files:
        exists = Path(path).exists()
        if skip_existing and exists:
            continue
        if exists and not confirm(f'{path} already exists, would you like to overwrite it?',
                                  default=False).unsafe_ask():
            continue
        with open(path, 'w') as file:
            file.write(data)

    print_success('Your model was created. Run `slate deploy` to deploy it to slate.')


def ensure_model(api: API):
    # create model if it doesn't exist
    try:
        with open('blankly.json', 'r') as file:
            data = json.load(file)
    except FileNotFoundError:
        print_failure('There was no model detected in this directory. Try running `slate init` to create one')
        sys.exit(1)

    if 'plan' not in data:
        data['plan'] = select('Select a plan:', [Choice(f'{name} - CPU: {info["cpu"]} RAM: {info["ram"]}', name)
                                                 for name, info in api.get_plans('live').items()]).unsafe_ask()

    if 'model_id' not in data or 'project_id' not in data:
        model = get_model_interactive(api)
        data['model_id'] = model['modelId']
        data['project_id'] = model['projectId']

    # save model_id and plan back into blankly.json
    with open('blankly.json', 'w') as file:
        json.dump(data, file, indent=4)

    return data


def missing_deployment_files(data) -> list:
    paths = [data['main_script'], 'slate.json', 'blankly.json', 'requirements.txt']
    return [path for path in paths if not Path(path).is_file()]


def slate_deploy(args):
    api = ensure_login()

    data = ensure_model(api)
    for path in missing_deployment_files(data):
        if not confirm(f'{path} is missing. Are you sure you want to continue?',
                       default=False).unsafe_ask():
            print_failure('Deployment cancelled')
            print_failure(f'You can try `slate init` to regenerate the {path} file.')
            return

    description = text('Enter a description for this version of the model:').unsafe_ask()

    with show_spinner('Uploading model') as spinner:
        model_path = zip_dir('.', data['ignore_files'])

        params = {
            'file_path': model_path,
            'project_id': data['project_id'],  # set by ensure_model
            'model_id': data['model_id'],  # set by ensure_model
            'version_description': description,
            'python_version': get_python_version(),
            'type_': data.get('type', 'strategy'),
            'plan': data['plan']  # set by ensure_model
        }
        if data.get('type', None) == 'screener':
            params['schedule'] = data['screener']['schedule']

        response = api.deploy(**params)
        if response.get('status', None) == 'success':
            spinner.ok('Model uploaded')
        else:
            spinner.fail('Error: ' + response['error'])


def get_model_interactive(api):
    create = select('Would you like to create a new model or attach to an existing one?',
                    [Choice('Create new model', True), Choice('Attach to existing model', False)]).unsafe_ask()
    if create:
        default_name = Path.cwd().name  # default name is working dir name
        name = text('Model name?', default=default_name, validate=validate_non_empty).unsafe_ask()
        description = text('Model description?', instruction='(Optional)').unsafe_ask()
        teams = api.list_teams()
        team_id = None
        if teams:
            team_choices = [Choice('Create on my personal account', False)] \
                           + [Choice(team.get('name', team['id']), team['id']) for team in teams]
            team_id = select('What team would you like to create this model under?', team_choices).unsafe_ask()
            return create_model(api, name, description, 'strategy', team_id or None)

    with show_spinner('Loading models...') as spinner:
        models = api.list_all_models()
        spinner.ok('Loaded')

    model = select('Select an existing model to attach to:',
                   [Choice(get_model_repr(model), model) for model in models]).unsafe_ask()
    return model


def get_model_repr(model: dict) -> str:
    name = model.get('name', model['id'])
    team = model.get('team', {}).get('name', None)
    if team:
        name = team + ' - ' + name
    return name


def generate_blankly_json(model: dict, script: str):
    data = {'main_script': script,
            'python_version': get_python_version(),
            'requirements': './requirements.txt',
            'working_directory': '.',
            'model_id': model['id'],
            'project_id': model['projectId'],
            'ignore_files': ['.git', '.idea', '.vscode']}
    return json.dumps(data, indent=4)


def generate_slate_json(model: dict, api: API):
    project_id = model['projectId']
    project_keys = api.generate_keys(project_id)
    data = {
        'api_key': project_keys['apiKey'],
        'api_pass': project_keys['apiPass'],
        'model_id': model['id'],
        'project_id': project_id
    }
    return json.dumps(data, indent=4)


def slate_login(args):
    if is_logged_in():
        print_success('You are already logged in')
        return

    launch_login_flow()


def slate_logout(args):
    with show_spinner('Logging out of Slate') as spinner:
        try:
            logout()
        except Exception:
            spinner.fail('Failed to logout')
            raise
        spinner.ok('Logged out')


def main():
    parser = argparse.ArgumentParser(description='Slate CLI')
    subparsers = parser.add_subparsers(required=True)

    init_parser = subparsers.add_parser('init', help='Initialize a slate model in the current directory')
    init_parser.set_defaults(func=slate_init)

    deploy_parser = subparsers.add_parser('deploy', help='Deploy a new version of your model to slate')
    deploy_parser.set_defaults(func=slate_deploy)

    login_parser = subparsers.add_parser('login', help='Login to Slate')
    login_parser.set_defaults(func=slate_login)

    logout_parser = subparsers.add_parser('logout', help='Logout of Slate')
    logout_parser.set_defaults(func=slate_logout)

    # run the selected command
    args = parser.parse_args()
    try:
        args.func(args)
    except KeyboardInterrupt:
        print_failure('Cancelled by user')
    except Exception:
        print_failure('An error occurred. Traceback:')
        traceback.print_exc()


if __name__ == '__main__':
    main()
